{-# LANGUAGE OverloadedStrings #-}

module UL0Parser where

import UntypedLambda

import Data.Attoparsec.Text

import Data.String (fromString)
import Data.Text (unpack)

expressionparser :: Parser Expression
expressionparser = choice [lambdaparser, applyparser, parensparser, identparser]

-- This function parses an expression enclosed in parentheses.
parensparser :: Parser Expression
parensparser = do
  char '('
  many' space
  exp <- expressionparser
  many' space
  char ')'
  return exp

-- this function parses a single identifier and wraps it into our datatypes.
identparser :: Parser Expression
identparser = Var . Identifier <$> many1 letter

--parses a lambda expression, like \x.x 
lambdaparser :: Parser Expression
lambdaparser = do
  choice [char '\\', char 'λ']
  ids <- sepBy1' (many1 letter) (many1' space)
  many' space
  char '.'
  many' space
  Lambda (fmap Identifier ids) <$> expressionparser


-- parses an application. 
-- It will first try to parse as many expressions separated by a space as possible.
-- To avoid an infinite loop, the sub-expression must not be applications themselves.
-- We use a helper function, nonrececpparse (non-recursive expression parser) to 
-- overcome this.
--
-- When we have a list of expression, since application is left associative
-- we run foldl1 to build our final expression, made of nested applications.
--
-- i.e.: a b c must be parsed exactly as (a b) c
applyparser :: Parser Expression
applyparser = do
  firstexp <- nonrecexpparse
  many1' space
  otherexp <- sepBy1' nonrecexpparse (many1' space)
  return $ leftassoc firstexp otherexp
  -- Application e1 <$> expressionparser
    where nonrecexpparse = choice [parensparser, identparser, lambdaparser]
          leftassoc :: Expression -> [Expression] -> Expression
          leftassoc x xs = foldl1 Application (x:xs)


-- parses a full statement. 
-- A statement is either an import, a definition, or an expression.
statementparser :: Parser Statement
statementparser = choice [importparser, defineparser, Exp <$> expressionparser]

-- parses a definition: an indetified followed by '=' and then an expression.
defineparser :: Parser Statement
defineparser = do
  id <- Identifier <$> many1 letter
  many' space
  char '='
  many' space
  Def id <$> expressionparser

-- parses an import statement. it just matches the import keyword, and then 
-- considers whatever's left on the line as the filename to attempt opening,
importparser :: Parser Statement
importparser = do
  string "import"
  many' space
  Imp . unpack <$> takeText

-- this is roughly taken from Attoparsec's documentation and it's how we 
-- actually run the parser. the fromString part is to convert a string to a bytestring.
runparse = parseOnly statementparser . fromString

