module UntypedLambda
(Environment, Statement(..), Expression(..), Identifier(..), updateEnv, envStrings, emptyEnv, normalform)

where

import qualified Data.Map as M
import Data.Bifunctor

-- Let's start defining a few wrapper types, togheter with
-- the definition of our lambda expression and a map-based environment.

-- IDENTIFIERS
newtype Identifier = Identifier String
  deriving (Eq, Ord)

instance Show Identifier where
  show (Identifier s) = s

-- THE ENVIRONMENT
newtype Environment = Env (M.Map Identifier Expression)



-- function we'll use to insert and update values in the environment.
updateEnv :: Environment -> Identifier -> Expression -> Environment
updateEnv (Env en) id exp = Env (M.insert id exp en)


envStrings (Env a) = map (bimap show show) (M.assocs a)

-- the empty environment. We need this to start our program.
emptyEnv :: Environment
emptyEnv = Env M.empty

lookupEnv :: Environment -> Identifier -> Maybe Expression
lookupEnv (Env en) = flip M.lookup en

-- EXPRESSION
data Expression = Var Identifier
                | Lambda [Identifier] Expression
                | Application Expression Expression
  deriving (Eq)

-- this is a big show instance, but it's mostly uninteresting code 
-- to pretty print our lambda expressions.
instance Show Expression where
  show (Var s) = show s
  show (Lambda ids ex) = "λ" ++ unwords (map show ids) ++ "." ++ show ex
  show (Application m n) = enclosem m ++ " " ++ enclosen n
    where
      enclosem k@(Lambda _ _) = enclose $ show k
      enclosem x = show x
      enclosen k@(Lambda _ _) = enclose $ show k
      enclosen k@(Application _ _) = enclose $ show k
      enclosen x = show x
      enclose s = "(" ++ s ++ ")"

-- STATEMENTS

-- a program in our language will be defined as a list of single-line statements, such as 

-- id = \x.x 
-- id y
-- import "FileName"
--
-- the first is a definition, and will update the environment.
-- the second is an expression that will be reduced to normal form and printed.

data Statement = Def Identifier Expression | Exp Expression | Imp String

-- now, the actual functions:

apply :: Environment -> Expression -> Expression -> Expression
apply en (Lambda [] ex) arg = reduce en arg
apply en (Lambda [i] ex) arg = reduce (updateEnv en i arg) ex
apply en (Lambda (i:ids) ex) arg = Lambda ids (reduce (updateEnv en i arg) ex)
apply en m n = Application (reduce en m) (reduce en n)

reduce :: Environment -> Expression ->  Expression
reduce en (Application m n) = apply en (reduce en m) n
reduce en (Lambda ids m) = Lambda ids (reduce en m)
reduce en (Var id) = case lookupEnv en id of
  Nothing -> Var id -- the variable is free, leave as it is.
  Just exp -> exp

tries :: Int
tries = 50

normalform :: Environment -> Expression -> Maybe Expression
normalform = loop tries
  where loop :: Int -> Environment -> Expression -> Maybe Expression
        loop 0 _ _ = Nothing
        loop n en ex = if reduce en ex == ex then Just ex else loop (n-1) en (reduce en ex)



